#ifndef XTHREADS_H
#define XTHREADS_H


#if defined(WIN32) || defined(__WIN32__)
#  include <windows.h>
#elif defined(MAC)
// JJ - No longer needed under Mach-O
//#include <Multiprocessing.h>
#elif defined(SMX)
#else
#  include <pthread.h>
#  ifndef sem_init
#    include <semaphore.h>
#  endif
#endif


#include "XTypes.h"
//#include "XContainers.h"




void Exit(void);


// This is an abstract interface that allows us to pass
// derived objects into XSmartLock
class IThreadLock
{
public:
	virtual ~IThreadLock(void) {}
	virtual void Lock(void) = 0;
	virtual void UnLock(void) = 0;
	virtual bool IsLocked(void) const = 0;
};

#if defined(WIN32)
	#define INF				0xFFFFFFFE
	typedef int				XThreadID;
	typedef HANDLE			XSemHandle;
	#define 	kInvalidThreadID -1

	typedef  ulong (__stdcall * XThreadProc)(void *arg);
#elif defined(SMX)
	#define INF				0xFFFFFFFF
	typedef	TCB_PTR			XThreadID;
	typedef SCB_PTR			XSemHandle;
	typedef void (*	XThreadProc)();
#elif defined (MAC)
	#define INF 			1247000
	typedef MPTaskID 		XThreadID;
	typedef MPSemaphoreID	XSemHandle;
	typedef TaskProc 		XThreadProc;
	#define kInvalidThreadID 0
#endif


extern 	XThreadID 		gMainThreadID;
extern 	XThreadID 		gAudioIOThreadID;
extern 	XThreadID 		gRecorderThreadID;
extern 	XThreadID 		gFileSchedulerThreadID;


// Platform idependent constants to tell us how
// a semaphore wait exited
enum
{
	X_WAIT_SIGNALLED			= 0,
	X_WAIT_ABANDONED			= 1,
	X_WAIT_TIMEOUT				= 2,
	X_WAIT_FAILED				= 3
};

enum
{
	X_PRIORITY_IDLE				=	0,
	X_PRIORITY_LOW				=	1,
	X_PRIORITY_MEDIUM			=	3,
	X_PRIORITY_HIGH				=	4,
	X_PRIORITY_REALTIME			=	5,
	X_PRIORITY_REALTIME_MINUS1	=	6			// Needed for Client/Server Audio Priorities (Server MUST Be higher priority)
};

#define X_PRIORITY_NORMAL			X_PRIORITY_MEDIUM

#define X_PRIORITY_DISPLAY		X_PRIORITY_NORMAL

long		AtomicInc(long *target);
long		AtomicDec(long *target);

bool		XThreadEqual(XThreadID id);
XThreadID	XThreadSelf(void);					// Get current thread
void		XSleep(uint32 milliseconds);
uint32		XGetTicks(void);


class XSemaphore : public XUnknown
{
public:
	XSemaphore(int initValue = 1);
	XSemaphore(const XSemaphore &theSource);
	virtual ~XSemaphore(void);

	void Signal(void);
	long Test(uint32 msTimeOut);

//	XSemaphore& operator=(XSemaphore &theSrc);

private:
	void			DeferedInit(void);		// We do not create the semaphore until it is first accessed (due to SMX limitations on Globals)

	XSemHandle		mySemHandle;
	int				myInitValue;
	bool			myOwnsSema;				// This is a copy flag for the operator= needed for MsgPorts
};

class XMutex : public IThreadLock
{
public:
	XMutex(void);
	virtual ~XMutex(void);

	void Init(void);

	void Lock(void);
	void UnLock(void);

	bool IsLocked(void) const { return (myLockCount > 0); }

private:
#if defined(WIN32) || defined(SMX)
    CRITICAL_SECTION	myCritSec;
#elif defined(MAC)
	MPCriticalRegionID	myMacID;
#else	// POSIX compliant
    pthread_mutex_t		myMtx;
#endif

	uint32				myLockCount;
	bool				myInited;
};

// JJ - I had to name this thread Event class
// with a 'T' since we already had a XEvent class (MIDI)
// This Thread sync object utilizes Win32 Events
// Not sure how we will implement this on MAC
// but at least we are platform safe
class XTEvent
{
public:
	XTEvent(void);
	virtual ~XTEvent(void);

	void Create(void);					// creates a local event
	void Create(const char *shName);	// creates an inter-process accessible event	

	long Wait(uint32 msTimeOut);
	void Set(void);

private:
#if defined(WIN32)
    HANDLE	myWin32Event;
#elif defined(MAC)
#else	// POSIX compliant
#endif
};


////////////////////////////////////////////////////////////////////////////////////
// A Stack based Locking Object - that locks/unlocks some IThreadLock
// objects automatically.
// BT: NOTE: We have to fiqure out how to pass in a Template like ObjectArray
// to a XSmartLock.

class XSmartLock
{
public:
	XSmartLock(IThreadLock *theObject);
	virtual ~XSmartLock(void);
	void Lock(void);
	void UnLock(void);

private:
	IThreadLock *myObject;
	bool		myOwnsLock;

	// These functions are unsupported....
	XSmartLock(void);
	XSmartLock(const XSmartLock &rhs);				// Copy constructor
	XSmartLock & operator= (const XSmartLock &rhs);	// assignment
};


////////////////////////////////////////////////////
// A class to store info on all Threads created
#define MAX_THREAD_NAME		128
class XThreadInfo
{
public:
	char		myName[MAX_THREAD_NAME];		// Should be long enough
	long		myPriority;
	XThreadID	myThreadID;
};

// This calls is there to monitor threads that are not created by us... ie. the ASIO I/O Thread
XThreadInfo * CreateThreadInfo(const char *threadName, XThreadID theID = kInvalidThreadID, long priority = X_PRIORITY_NORMAL);


// Hopefully you will choose to use the NEW thread class below instead
// of using these functions..however they are here just incase.
XThreadID	XCreateThread(XThreadProc proc, long priority, long argument, const char *threadName, uint32 affinityMask = 0xffffffff);
XErr		XDeleteThread(XThreadID threadID);


class XThread: public XUnknown
{
public:
    XThread(bool isAutoFree, const char *name);
    virtual ~XThread();

#ifdef WIN32
    XThreadID  GetId(void)   { return XThreadID(myId); }
#else
    XThreadID  GetId(void)   { return myHandle; }
#endif

    bool IsRunning(void)	{ return myRunning != 0; }
    bool IsFinished(void)	{ return myFinished; }
    bool IsSignaled(void)	{ return mySignaled; }

    void Start(long priority);
    void Signal(void);
    void WaitFor(void);

	// Util functions for getting info on all Threads
	static uint32 CountThreads(void);
	static XThreadInfo *GetThreadInfo(uint32 threadIndex);

protected:
#ifdef WIN32
    ulong myId;
#endif
    XThreadID myHandle;
    bool myAutoFree;
    int  myRunning;
    bool myFinished;
    bool mySignaled;
	XThreadInfo *myThreadInfo;

	// Override these for your particular thread class
    virtual void Execute(void) = 0;	
    virtual void Cleanup(void) = 0;

    friend void _ThreadEpilog(XThread* thr);

#ifdef WIN32
    friend ulong __stdcall _ThreadProc(void* arg);
#else
    friend long _ThreadProc(void* arg);
#endif

	friend XThreadID	XCreateThread(XThreadProc proc, long priority, long argument, const char *threadName, uint32 affinityMask);
	friend void			CheckMain(void);
};



#endif		// XTHREADS_H

